/*
 * Estonian ID card plugin for web browsers
 *
 * Copyright (C) 2010-2011 Codeborne <info@codeborne.com>
 *
 * This is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

Components.utils.import("resource://esteid-pkcs11-module-loader/global.js");

const EstEidModName = "Estonian ID Card";
const nsModDB = "@mozilla.org/security/pkcs11moduledb;1";
const nsHttpProtocolHandler = "@mozilla.org/network/protocol;1?name=http";
const nsLocalFile = "@mozilla.org/file/local;1";
const nsPKCS11 = "@mozilla.org/security/pkcs11;1";

const dllName = "onepin-opensc-pkcs11";

var Ci = Components.interfaces;
var Cc = Components.classes;

var certDB = Cc["@mozilla.org/security/x509certdb;1"].getService(Ci.nsIX509CertDB);

function log(message) {
  if (console) {
    console.log(message);
  }
}

function removeModule(module) {
  var pkcs11 = Cc[nsPKCS11].getService(Ci.nsIPKCS11);
  try {
    pkcs11.deleteModule(module);
  }
  catch(e) {
    log("Unable to remove module " + module + ":" + e);
  }
}

function loadModule(moduleDll) {
  var pkcs11 = Cc[nsPKCS11].getService(Ci.nsIPKCS11);
  try {
    pkcs11.addModule(EstEidModName, moduleDll, 0x1 << 28, 0);
  }
  catch(e) {
    log("Unable to load module: " + e);
  }
}
function ModuleLoader() {
  var platform = detectOS();
  var bits = detectArchitecture();

  var moduleDll;

  if (platform == "Macintosh") {
    moduleDll = "/Library/OpenSC/lib/" + dllName + ".so";
  } else if (platform == "X11") {
    var libDir = (bits == 64) ? "/usr/lib/x86_64-linux-gnu/" : "/usr/lib/i386-linux-gnu/";
    moduleDll = libDir + dllName + ".so";
  } else if (platform == "Windows") {
    moduleDll = dllName + ".dll";
  }
  else {
    throw "Failed to detect Operating System."
  }

  var securityModuleDb = Cc[nsModDB].createInstance(Ci.nsIPKCS11ModuleDB);
  var modules = securityModuleDb.listModules();

  var needToLoad = true;
  for (; ;) {
    var module = modules.currentItem().QueryInterface(Ci.nsIPKCS11Module);
    /* Remove modules that have recognized names but different DLLs */
    if (module) {
      if (module.libName == moduleDll && module.name == EstEidModName) {
        needToLoad = false;
      } else if (module.name == EstEidModName || /(opensc-pkcs11|esteid-pkcs11)\.(so|dll)$/.test(module.libName)) {
        removeModule(module)
      }
    }
    try {
      modules.next();
    }
    catch(e) {
      break;
    }
  }

  if (needToLoad) {
    loadModule(moduleDll);
  }
}

function FindCertificateByNickname(x509CertNickName) {
  var certFound = true;
  try {
    var x509cert = certDB.findCertByNickname(null, x509CertNickName);
  }
  catch(err) {
    certFound = false;
  }
  return certFound;
}

function installCertificateIfNotExist(certName) {
  if (!FindCertificateByNickname(certName + " - AS Sertifitseerimiskeskus")) {
    return addCertificate(certName + ".crt", 'C,c,c');
  }
  return false;
}

function CertificateImporter() {
  var restartNeeded = false;

  if (!FindCertificateByNickname("ESTEID-SK - AS Sertifitseerimiskeskus")) {
    if(!FindCertificateByNickname("SK-ESTEID")){
      restartNeeded = addCertificate("ESTEID-SK.crt", 'C,c,c');
    }
  }
  if (!FindCertificateByNickname("ESTEID-SK 2007 - AS Sertifitseerimiskeskus")) {
    if(!FindCertificateByNickname("SK-ESTEID-2007")){
      restartNeeded = addCertificate("ESTEID-SK 2007.crt", 'C,c,c');
    }
  }

  if (!FindCertificateByNickname("Builtin Object Token:EE Certification Centre Root CA") &&
      !FindCertificateByNickname("EE Certification Centre Root CA - AS Sertifitseerimiskeskus") &&
      !FindCertificateByNickname("EE Certification Centre Root CA")){
        restartNeeded = addCertificate("EECCRCA.crt", 'C,c,c');
  }

  if (!FindCertificateByNickname("Builtin Object Token:Juur-SK") &&
      !FindCertificateByNickname("Juur-SK - AS Sertifitseerimiskeskus") &&
      !FindCertificateByNickname("Juur-SK")){
        restartNeeded = addCertificate("JUUR-SK.crt", 'C,c,c');
  }

  var newCertificates=["EID-SK 2011", "ESTEID-SK 2011"];
  for (var i = 0; i < newCertificates.length; i++) {
    var result = installCertificateIfNotExist(newCertificates[i]);
    restartNeeded = (restartNeeded || result);
  }

  if (restartNeeded) restartFF();
}

function restartFF() {
  var boot = Cc["@mozilla.org/toolkit/app-startup;1"].getService(Ci.nsIAppStartup);
  boot.quit(Ci.nsIAppStartup.eForceQuit | Ci.nsIAppStartup.eRestart);
}

function addCertificate(CertName, CertTrust) {
  var file = Cc["@mozilla.org/file/local;1"].createInstance(Ci.nsILocalFile);
  try {
    log(CertName);
    file.initWithPath(getCertPath() + CertName);

    var fileStream = Cc["@mozilla.org/network/file-input-stream;1"].createInstance(Ci.nsIFileInputStream);
    fileStream.init(file, 1, 0, false);
    var scriptableStream = Cc["@mozilla.org/scriptableinputstream;1"].getService(Ci.nsIScriptableInputStream);
    scriptableStream.init(fileStream);
    var certfile = scriptableStream.read(scriptableStream.available());
    scriptableStream.close();
    fileStream.close();

    var beginCert = "-----BEGIN CERTIFICATE-----";
    var endCert = "-----END CERTIFICATE-----";

    var certDB = Cc["@mozilla.org/security/x509certdb;1"].getService(Ci.nsIX509CertDB2);

    certfile = certfile.replace(/[\r\n]/g, "");
    var begin = certfile.indexOf(beginCert);
    var end = certfile.indexOf(endCert);
    var cert = certfile.substring(begin + beginCert.length, end);
    certDB.addCertFromBase64(cert, CertTrust, "");

    return true;
  }
  catch(err){
    log(err);
    return false;
  }
}

function detectOS() {
  var httpHandler = Cc[nsHttpProtocolHandler].getService(Ci.nsIHttpProtocolHandler);
  return httpHandler.platform;
}

function detectArchitecture() {
  var httpHandler = Cc[nsHttpProtocolHandler].getService(Ci.nsIHttpProtocolHandler);
  var osCPU = httpHandler.oscpu;
  return (osCPU.indexOf(" x86_64") < 0 && osCPU.indexOf(" x64") < 0) ? 32 : 64;
}

function getCertPath() {
  var platform = detectOS();
  var bits = detectArchitecture();

  if (platform == "Macintosh") {
    return "/Library/EstonianIDCard/config/certs/";
  }
  else if (platform == "X11"){
    return "/usr/share/esteid/certs/";
  }
  else if (platform == "Windows"){
    var certsLocation;
    try {
      var key = Cc["@mozilla.org/windows-registry-key;1"].createInstance(Ci.nsIWindowsRegKey);
      key.open(Ci.nsIWindowsRegKey.ROOT_KEY_LOCAL_MACHINE, "Software\\Estonian ID Card", Ci.nsIWindowsRegKey.ACCESS_READ);
      certsLocation = key.readStringValue("Installed") + "\\certs\\";
    } catch(e) {
      certsLocation = "C:\\Program Files" + ((bits == 64) ? " (x86)" : "") + "\\Estonian ID Card\\certs\\";
    }
    return certsLocation;
  }
  else{
    throw("Unable detect operating system. Loading certificates is cancelled");
  }
}

try {
  if (!loaded) {
    ModuleLoader();
    CertificateImporter();
  }
  loaded = true;
}
catch (e) {
  log(e)
}
